#! /usr/bin/env python
# -*- coding: latin-1 -*-

"""Quick-quick implementation of WikiWikiWeb in Python
"""

# Copyright (C) 1999, 2000 Martin Pool <mbp@humbug.org.au>
# Changes (c) 2003, 2004 Panu Kalliokoski <atehwa@sange.fi>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA


import cgi, sys, string, os, re, errno, time, stat
from fs import FileSystem
from cgi import log
from os import path, environ, umask
from socket import gethostbyaddr
from time import localtime, strftime
from cStringIO import StringIO
from urllib import quote as urlquote, unquote as urlunquote
htmlquote = cgi.escape
from sets import Set

# Configurable parts ------------------------------------------
data_dir = '/tmp/wiki-test' # all data under this directory
sys.stderr = open(path.join(data_dir, 'mbp_piki_err'), 'at')
cgi.logfile = path.join(data_dir, 'cgi_log')

text_dir = path.join(data_dir, 'fs') # the actual pages here
editlog_name = path.join(data_dir, 'editlog')
logo_string = 'PikiPiki '
changed_time_fmt = '[%I:%M %p]'
date_fmt = '%a %d %b %Y'
datetime_fmt = '%a %d %b %Y %I:%M %p'
show_hosts = 0                          # show hostnames?
css_url = ''                            # stylesheet link, or ''
nonexist_qm = 1                         # show '?' for nonexistent?

# What kind of words are allowed in queries.
word_re_str = r"[A-Za-z-0-9 _.,+'!#%&=?;:><]+"
word_anchored_re = re.compile('^' + word_re_str + '$')
command_re_str = "(search|edit|fullsearch|titlesearch)\=(.*)"
link_re = re.compile(r"(http|ftp|mailto|nntp|news):")

# Editlog -----------------------------------------------------------
# Functions to keep track of when people have changed pages, so we can
# do the recent changes page and so on.
# The editlog is stored with one record per line, as tab-separated
# words: page_name, host, time

def editlog_add(page_name, host):
    editlog = open(editlog_name, 'a+')
    try: 
        # fcntl.flock(editlog.fileno(), fcntl.LOCK_EX)
        editlog.seek(0, 2)                  # to end
        editlog.write(string.join((page_name, host, `time.time()`), "\t") + "\n")
    finally:
        # fcntl.flock(editlog.fileno(), fcntl.LOCK_UN)
        editlog.close()


def editlog_raw_lines():
    editlog = open(editlog_name, 'rt')
    try:
        # fcntl.flock(editlog.fileno(), fcntl.LOCK_SH)
        return editlog.readlines()
    finally:
        # fcntl.flock(editlog.fileno(), fcntl.LOCK_UN)
        editlog.close()



# Formatting stuff --------------------------------------------------
# These are needed from multiple places and so are gathered here.

def emit_header():
    print "Content-type: text/html; charset=ISO-8859-1"
    print


def get_scriptname():
    try: return "http://%s%s/" % \
        (environ['SERVER_NAME'], environ['SCRIPT_NAME'])
    except KeyError: return ''


def send_title(text, link=None, msg=None):
    emit_header()
    print "<head><title>%s</title>" % text
    print """<link rel="alternate" type="application/rss+xml" href="%s?rss=rss">""" % get_scriptname()
    print """<style type="text/css">
    BODY { background-color: #FFFFFF; color: #000000; }
    A { color: #1f6b9e }
    A.nonexistent { background-color: #CCCCCC }
    DIV.navbar { float: right; width: 11em; padding: 10px; font-size: smaller; }
    </style>"""
    print "</head>"
    print '<body><h2>'
    if logo_string:
        print link_tag('RecentChanges', logo_string)
    if link:
        print link_tag(link, text)
    else:
        print text
    print '</h2>'
    if msg:
        print msg
        print '<hr>'


def link_tag(params, text=None, ss_class=None):
    if text is None:
        text = params           # default
    if ss_class:
        classattr = 'class="%s" ' % ss_class
    else:
        classattr = ''
    if not len(environ.get('PATH_INFO', '')):
        params = get_scriptname() + params
    return '<a %shref="%s">%s</a>' % (classattr, params, text)


# Actions ---------------------------------------------------
# These are stuff directed by QUERY_STRING arguments.  They are
# used for most everything except normal page browsing.

def do_fullsearch(needle):
    send_title('Searched for "%s" from page contents' % (needle))

    needle_re = re.compile(needle, re.IGNORECASE)
    hits = []
    all_pages = page_list()
    for page_name in all_pages:
        body = Page(page_name).get_raw_body()
        count = len(needle_re.findall(body))
        if count:
            hits.append((count, page_name))

    # The default comparison for tuples compares elements in order,
    # so this sorts by number of hits
    hits.sort()
    hits.reverse()

    print "<UL>"
    for (count, page_name) in hits:
        print '<LI>' + Page(page_name).link_to()
        print ' . . . . ' + `count`
        print ['match', 'matches'][count <> 1]
    print "</UL>"

    print_search_stats(len(hits), len(all_pages))


def do_titlesearch(needle):
    # TODO: check needle is legal -- but probably we can accept any RE

    send_title('Search for "%s" from page titles' % needle)
    
    needle_re = re.compile(needle, re.IGNORECASE)
    all_pages = page_list()
    hits = filter(needle_re.search, all_pages)

    print "<UL>"
    for filename in hits:
        print '<LI>' + Page(filename).link_to()
    print "</UL>"

    print_search_stats(len(hits), len(all_pages))


def print_search_stats(hits, searched):
    print "<p>%d found " % hits
    print " from a total of %d pages." % searched


def do_edit(pagename):
    Page(pagename).send_editor()


def do_rss(dummy):
    print 'Content-type: text/xml; charset=ISO-8859-1'
    print
    print '<?xml version="1.0" encoding="ISO-8859-1"?>'
    print '<rss version="0.91"><channel>'
    print '<title>%s</title>' % logo_string
    print '<link>%s</link>' % get_scriptname()
    print '<description>RSS of recent changes in %s</description>' % \
                logo_string
    print publish_change_summary('rss')
    print '</channel></rss>'


def no_nonempty_common_lines(t1, t2):
    common_lines = Set(t1.split('\n')).intersection(Set(t2.split('\n')))
    nonempty_common_lines = filter(lambda x: x, common_lines)
    return not nonempty_common_lines

def too_much_new_links(t1, t2):
    new_lines = Set(t1.split('\n')).difference(Set(t2.split('\n')))
    count = 0
    for line in new_lines:
        count = count + len(link_re.findall(line))
    return count > 2

def do_savepage(pagename):
    global form
    status_map = { None: 'no', False: 'ok', True: 'confl' }
    def norm_ws( string ):
        return string.replace( '\r\n', '\n' ).rstrip() + '\n'
    pg = Page(pagename)
    problem = None
    try:
        savetext = norm_ws( form['savetext'].value )
        oldtext = norm_ws( urlunquote( form['oldtext'].value ))
    except KeyError: problem = "Fix your calibration."
    else:
        if no_nonempty_common_lines(savetext, oldtext):
            problem = "Too painful change."
        if too_much_new_links(savetext, oldtext):
            problem = "Don't be quite so pushy."
    if problem:
        emit_header()
        print "<html><h1>Boo! %s</h1></html>" % problem
        return

    save_status = status_map[pg.save_text( savetext, oldtext )]

    redir_url = '%s%s?saved=%s' % \
            (get_scriptname(), urlquote( pagename ), save_status)
    emit_header()
    print """<html><head><title>Saved %s</title>
        <meta http-equiv="Refresh" content="0; url=%s"></head>
        <body><h1>Your changes were saved.</h1> Look
        <a href="%s">here</a>""" % (pagename, redir_url, redir_url)


def make_index_key():
    s = '<p><center>'
    links = map(lambda ch: '<a href="#%s">%s</a>' % (ch, ch),
                string.lowercase)
    s = s + string.join(links, ' | ')
    s = s + '</center><p>'
    return s


def page_list():
    return fs.keys()


def print_footer(name, editable=1, mod_string=None):
    print '<hr>'
    if editable:
        print link_tag('?edit='+urlquote(name), 'Edit')
        print "this page"
        if mod_string:
            print "(last modified %s)" % mod_string
        print '<br>'
    print link_tag('FindPage?value='+urlquote(name), 'Search')
    print " for a page"


# Macros ----------------------------------------------------------
# These are special markers embedded in the page markup that do
# complicated and wondrous things.

def _macro_titlesearch():
    return _macro_search("titlesearch")

def _macro_fullsearch():
    return _macro_search("fullsearch")

def _macro_search(type):
    if form.has_key('value'):
        default = form["value"].value
    else:
        default = ''
    return """<form method=get>
    <input name=%s size=30 value="%s"> 
    <input type=submit value="Go">
    </form>""" % (type, default)

def _macro_WantedPages():
    s = "<ul>\n"
    link_re = re.compile( r"\[(" + word_re_str + r")\]" )
    wanted = {}
    for page in page_list():
        p = Page( page )
        body = p.get_raw_body()
        for match in link_re.findall( body ):
            name = match.lower()
            plink = p.link_to()
            try: wanted[name].append( plink )
            except KeyError:
                if Page( name ).exists(): continue
                wanted[name] = [plink]
    for (name, reflist) in wanted.items():
        s = s + "<li> %s (%s) </li>\n" % \
            (Page( name ).link_to(), '; '.join( reflist ))
    return s + "</ul>\n"

def _macro_goto():
    return """<form method=get><input name=goto size=30>
    <input type=submit value="Go">
    </form>"""

def _macro_titleindex():
    s = make_index_key()
    pages = list(page_list())
    pages.sort()
    current_letter = None
    for name in pages:
        letter = string.lower(name[0])
        if letter <> current_letter:
            s = s + '<a name="%s"><h3>%s</h3></a>' % (letter, letter)
            current_letter = letter
        else:
            s = s + '<br>'
        s = s + Page(name).link_to()
    return s


def _macro_recentchanges():
    return link_tag( '?rss=rss', 'Changes as RSS' ) + '\n' + \
        publish_change_summary('html')


def publish_change_summary(format):
    lines = editlog_raw_lines()
    lines.reverse()
    
    ratchet_day = None
    done_words = {}
    item_i = 0
    buf = StringIO()
    for line in lines:
        page_name, addr, ed_time = string.split(line, '\t')

        if done_words.has_key(page_name.lower()):
            continue

        item_i = item_i + 1
        if item_i > { 'html':50, 'rss':10 }[format]: break

        # year, month, day, DoW
        time_tuple = localtime(float(ed_time))
        day = tuple(time_tuple[0:3])
        if day <> ratchet_day and format == 'html':
            buf.write('<h3>%s</h3>' % strftime(date_fmt, time_tuple))
            ratchet_day = day

        done_words[page_name.lower()] = 1

        if format == 'html':
            buf.write(Page(page_name).link_to())
            if show_hosts:
                buf.write(' . . . . ')
                try:
                    buf.write(gethostbyaddr(addr)[0])
                except:
                    buf.write(str(addr))
            changetime = time.strftime(changed_time_fmt, time_tuple)
            buf.write(' . . . . ')
            buf.write(changetime)
            buf.write('<br>\n')

        elif format == 'rss':
            buf.write(Page(page_name).rss_item())

    return buf.getvalue()

macros = {
        'TitleSearch' : _macro_titlesearch,
        'FullSearch' : _macro_fullsearch,
        'WantedPages' : _macro_wantedpages,
        'GoTo' : _macro_goto,
        'TitleIndex' : _macro_titleindex,
        'RecentChanges' : _macro_recentchanges
}

# ----------------------------------------------------------
class PageFormatter:
    """Object that turns Wiki markup into HTML.

    All formatting commands can be parsed one line at a time, though
    some state is carried over between lines.
    """
    def __init__(self, raw):
        self.raw = raw
        self.is_em = self.is_b = 0
        self.blocktype = ''
        self.in_pre = 0
        self.headlines = []

    def print_toc(self):
        if not self.headlines: return
        print '<a name="toc"></a>'
        print '<hr><p>Shortcuts:\n<ul>'
        for i in range(len(self.headlines)):
            print '<li><a href="#hdl%d">%s</a></li>' % \
                ( i, self.headlines[i] )
        print '</ul>'


    def set_blocktype(self, blocktype):
        res = self.close_inlines()
        if blocktype == self.blocktype: return res
        return res + self.end_block() + self.start_block(blocktype)

    def close_inlines(self):
        res = ['', '</em>'][self.is_em] + ['', '</b>'][self.is_b]
        self.is_em = self.is_b = 0
        return res

    def start_block(self, block):
        self.blocktype = block
        if not block: return ''
        return '<%s>' % block

    def end_block(self):
        if not self.blocktype: return ''
        return '</%s>' % self.blocktype


    def _emph_repl(self, word):
        if self.in_pre: return word
        if len(word) == 3:
            self.is_b = not self.is_b
            return ['</b>', '<b>'][self.is_b]
        else:
            self.is_em = not self.is_em
            return ['</em>', '<em>'][self.is_em]

    def _rule_repl(self, word):
        if self.in_pre: return word
        s = self.set_blocktype('')
        if len(word) <= 4:
            s = s + "\n<hr>\n"
        else:
            s = s + "\n<hr size=%d>\n" % (len(word) - 2 )
        return s

    def _word_repl(self, word):
        if self.in_pre: return word
        return Page(word).link_to()

    def _url_repl(self, word):
        return '<a href="%s">%s</a>' % (word, word)

    def _email_repl(self, word):
        return '<a href="mailto:%s">%s</a>' % (word, word)

    def _ent_repl(self, s):
        return {'&': '&amp;',
                '<': '&lt;',
                '>': '&gt;',
                '|]': ']'}[s]
    
    def _uli_repl(self, match):
        if self.in_pre: return match
        return self.set_blocktype('ul') + '<li>'

    def _oli_repl(self, match):
        if self.in_pre: return match
        return self.set_blocktype('ol') + '<li>'

    def _para_repl(self, match):
        if self.in_pre: return ''
        return self.set_blocktype('') + '<p>'

    def _headl_repl(self, line):
        if self.in_pre: return line
        text = line[3:]
        num = len(self.headlines)
        self.headlines.append(text)
        return self.set_blocktype('') + '''<a name="hdl%d"></a>
            <h3><a href="#toc">%s</a></h3>''' % ( num, text )

    def _pre_repl(self, word):
        if word == '{{{':
            self.in_pre = 1
            return self.set_blocktype('pre')
        else:
            self.in_pre = 0
            return self.set_blocktype('')

    def _macro_repl(self, word):
        macro_name = word[2:-2]
        try: return macros[macro_name]()
        except KeyError: return "(Unknown macro %s)" % macro_name

    def _link_repl(self, word):
        if self.in_pre: return word
        link_word = word[1:-1]
        return Page(link_word).link_to()

    def _br_repl(self, match):
        if self.in_pre: return word
        return '<br>'

    def replace(self, match):
        match_r = match.groupdict().items()
        for type, hit in match_r:
            if hit != None:
                return apply(getattr(self, '_' + type + '_repl'), (hit,))
        else:
            raise "Can't handle match " + `match_r`
        

    def print_html(self):
        # For each line, we scan through looking for magic
        # strings, outputting verbatim any intervening text
        scan_re = re.compile(
            r"(?:(?P<emph>'''?)"
            + r"|(?P<ent>[<>&]|\|\])"
            + r"|(?P<word>\b(?:[A-Z][a-z]+){2,}\b)"
            + r"|(?P<rule>^-{4,}$)"
            + r"|(?P<br>//\s*$)"
            + r"|(?P<url>(http|ftp|nntp|news|mailto)\:[^\s')(\"]+)"
            + r"|(?P<email>[-\w._+]+\@[\w.-]+)"
            + r"|(?P<uli>^\s+\*)"
            + r"|(?P<oli>^\s+#)"
            + r"|(?P<para>^\s*$)"
            + r"|(?P<headl>^!!!.*$)"
            + r"|(?P<pre>(\{\{\{|\}\}\}))"
            + r"|(?P<macro>\[\[[A-Za-z]*\]\])"
            + r"|(?P<link>\[" + word_re_str + r"\])"
            + r")")
        eol_re = re.compile(r'\r?\n')
        raw = string.expandtabs(self.raw)
        for line in eol_re.split(raw):
            print re.sub(scan_re, self.replace, line)
        if self.in_pre: print '</pre>'
        print self.set_blocktype('')
        self.print_toc()


def print_formatted(pagename):
    PageFormatter(Page(pagename).get_raw_body()).print_html()

# ----------------------------------------------------------
class Page:
    def __init__(self, page_name):
        word_match = re.match(word_re_str, page_name)
        if word_match: self.page_name = word_match.group(0)
        else: self.page_name = 'hacking'
        self.f = fs[self.page_name.lower()]

    def split_title(self):
        # look for the end of words and the start of a new word,
        # and insert a space there
        # return re.sub('([a-z])([A-Z])', r'\1 \2', self.page_name)
        return self.page_name[0].upper() + self.page_name[1:]

    def exists(self):
        return not not self.f

    def link_to(self):
        word = self.page_name
        if self.exists():
            return link_tag(urlquote(word), word)
        else:
            if nonexist_qm:
                return "%s[%s]" % (word,
                        link_tag('?edit=' + urlquote(word), '?', 'nonexistent'))
            else:
                return link_tag(urlquote(word), word, 'nonexistent')

    def rss_item(self):
        word = self.page_name
        mtime = self.f.modtime()
        if mtime:
            mdate = strftime( date_fmt, localtime( mtime ))
            mtime = strftime( changed_time_fmt, localtime( mtime ))
        else: mdate, mtime = '?', ''
        return ('<item><title>%s - %s</title>\n<link>%s%s</link>\n' +
                '<description>Changed %s:\n%s</description>\n</item>\n') % \
                ( word, mdate, get_scriptname(), urlquote(word),
                  mtime, self.diffs() )

    def diffs(self):
        diff = filter( lambda s: s[0] in '+@', self.f.file_diffs() )
        res = ''.join( map( htmlquote, diff[1:25] ))
        if len(diff) > 25: res = res + '...'
        return res

    def get_raw_body(self):
        if self.exists():
            return self.f.contents()
        else:
            return 'Please explain %s here.' % self.page_name

    def send_menu(self):
        print '<div class=navbar>'
        print_formatted('menu')
        print '</div>'

    def send_page(self):
        link = '?fullsearch=' + urlquote( self.page_name )
        if form.has_key( 'saved' ) and form['saved'].value != 'no':
            msg = "<p> <b>Thank you for participation in this wiki.</b> </p>"
            if form['saved'].value == 'confl': 
                msg = msg + "<p> (there was an editing conflict) </p>"
        else: msg = None
        send_title(self.split_title(), link, msg)
        self.send_menu()
        PageFormatter(self.get_raw_body()).print_html()
        print_footer(self.page_name, 1, self._last_modified())

    def _last_modified(self):
        mtime = self.f.modtime()
        if not mtime: return None
        return strftime( datetime_fmt, localtime( mtime ))

    def send_editor(self):
        send_title('Editing ' + self.split_title())
        print_formatted('editor-pre')
        print '<form method="post" action="%s">' % (get_scriptname())
        print '<input type=hidden name="savepage" value="%s">' % (self.page_name)
        raw_body = string.replace(self.get_raw_body(), '\r\n', '\n')
        print """<textarea wrap="virtual" name="savetext" rows="17"
                 cols="80">%s</textarea>""" % htmlquote( raw_body )
        print """<br><input type=submit value="Save">
                 <input type=reset value="Reset">
                 """
        print '<input type=hidden name="oldtext" value="%s">' % urlquote( raw_body )
        print "</form>"
        print_formatted('editor-post')
        print_footer(self.page_name, 0)

    def save_text(self, newtext, oldtext):
        umask(000)
        res = self.f.mergechanges( newtext, oldtext )
        remote_name = environ.get('REMOTE_ADDR', '')
        editlog_add(self.page_name, remote_name)
        return res


# ---- MAIN ----

try:
    fs = FileSystem( text_dir )
    form = cgi.FieldStorage()

    handlers = { 'fullsearch':  do_fullsearch,
                 'titlesearch': do_titlesearch,
                 'edit':        do_edit,
                 'rss':         do_rss,
                 'savepage':    do_savepage }

    for cmd in handlers.keys():
        if form.has_key(cmd):
            apply(handlers[cmd], (form[cmd].value,))
            break
    else:
        path_info = environ.get('PATH_INFO', '')

        if form.has_key('goto'):
            query = form['goto'].value
        elif len(path_info) and path_info[0] == '/':
            query = urlunquote( path_info[1:] ) or 'FrontPage'
        else:       
            query = urlunquote( environ.get('QUERY_STRING', '')) or 'FrontPage'

        word_match = re.match(word_re_str, query)
        if word_match:
            word = word_match.group(0)
            Page(word).send_page()
        else:
            print "<p>Can't work out query \"<pre>" + query + "</pre>\""

except:
    print
    print "<b>There was an internal error.</b>  Go check the error log for details."
    raise
    # cgi.print_exception()

sys.stdout.flush()

# vim:tabstop=8:shiftwidth=4:expandtab: 

